#include <string.h>
#include "gps.h"
#include "queue.h"

SerialGPS::SerialGPS(PinName tx, PinName rx, int baud)
{
#ifndef USE_ARDUINO
    ser(tx, rx);
    ser.baud(baud);
#else  // USE_ARDUINO
    Serial1.begin(baud);
#endif // USE_ARDUINO
    Init();
    gps_callback_table = NULL;
}

SerialGPS::~SerialGPS()
{
}

void SerialGPS::Init()
{
    queue.Init();
    SetRcvIsrState(RCV_ISR_IDLE);
}

void SerialGPS::Processing(void)
{
    while(ProcessingOneShot() == true);
}

bool SerialGPS::ProcessingOneShot(void)
{
    QueueData *queue_data;
    int32_t parse_result;
    
    // Dequeue
    if(queue.Dequeue(&queue_data) == false){
        return false;
    }

    //
    if(gps_callback_table != NULL){
        // Log Callback (Raw text NEMA sentence)
        if(gps_callback_table->log_cbfunc != NULL){
            gps_callback_table->log_cbfunc(queue_data->msg);
        }

        // Parse of nmea sentences
        if((gps_callback_table->gga_cbfunc != NULL) &&
           (strncmp((const char *)queue_data->msg, GPS_NMEA_GPGGA, GPS_NMEA_LEN) == 0)){
            // Parse of GGA sentence
            parse_result = ParseGGA(&queue_data->msg[6], queue_data->len-6, &nmea_gga);
            // Callback
            if((parse_result == 0) && (nmea_gga.error == 0)){
                gps_callback_table->gga_cbfunc(&nmea_gga);
            }
        } else if((gps_callback_table->gll_cbfunc != NULL) &&
                (strncmp((const char *)queue_data->msg, GPS_NMEA_GPGLL, GPS_NMEA_LEN) == 0)){
#ifdef GPS_NMEA_RESERVE
            // adding GLL sentence process.
            // Callback
            if((parse_result == 0) && (nmea_gll.error == 0)){
                gps_callback_table->gll_cbfunc(&nmea_gll);
            }
#endif
        } else if((gps_callback_table->gsa_cbfunc != NULL) &&
                (strncmp((const char *)queue_data->msg, GPS_NMEA_GPGSA, GPS_NMEA_LEN) == 0)){
#ifdef GPS_NMEA_RESERVE
            // adding GSA sentence process.
            // Callback
            if((parse_result == 0) && (nmea_gsa.error == 0)){
                gps_callback_table->gsa_cbfunc(&nmea_gsa);
            }
#endif
        } else if((gps_callback_table->gsv_cbfunc != NULL) &&
                  (strncmp((const char *)queue_data->msg, GPS_NMEA_GPGSV, GPS_NMEA_LEN) == 0)){
#ifdef GPS_NMEA_RESERVE
            // adding GSV sentence process.
            // Callback
            if((parse_result == 0) && (nmea_gsv.error == 0)){
                gps_callback_table->gsv_cbfunc(&nmea_gsv);
            }
#endif
        } else if((gps_callback_table->rmc_cbfunc != NULL) &&
                  (strncmp((const char *)queue_data->msg, GPS_NMEA_GPRMC, GPS_NMEA_LEN) == 0)){
#ifdef GPS_NMEA_RESERVE
            // adding RMC sentence process.
            // Callback
            if((parse_result == 0) && (nmea_rmc.error == 0)){
                gps_callback_table->rmc_cbfunc(&nmea_rmc);
            }
#endif
        } else if((gps_callback_table->vtg_cbfunc != NULL) &&
                  (strncmp((const char *)queue_data->msg, GPS_NMEA_GPVTG, GPS_NMEA_LEN) == 0)){
#ifdef GPS_NMEA_RESERVE
            // adding VTG sentence process.
            // Callback
            if((parse_result == 0) && (nmea_vtg.error == 0)){
                gps_callback_table->vtg_cbfunc(&nmea_vtg);
            }
#endif
        } else if((gps_callback_table->zda_cbfunc != NULL) &&
                (strncmp((const char *)queue_data->msg, GPS_NMEA_GPZDA, GPS_NMEA_LEN) == 0)){
            // Parse of ZDA sentence
            parse_result = ParseZDA(&queue_data->msg[6], queue_data->len-6, &nmea_zda);
            // Callback
            if((parse_result == 0) && (nmea_zda.error == 0)){
                gps_callback_table->zda_cbfunc(&nmea_zda);
            }
        } else {
            // (!!error!!)
            // adding illegal sentences process.
        }
    }
    // Comptleted queue process
    queue.DequeueComplete();
    
    return true;
}

void SerialGPS::SetGpsCallback(SerialGPS::GpsCallbackTable *table)
{
    gps_callback_table = table;
}

void SerialGPS::RcvIsrEnable(void)
{
    Init();
#ifndef USE_ARDUINO
    ser.attach(this, &SerialGPS::RcvIsrHandler);
#else  // USE_ARDUINO
#endif // USE_ARDUINO
}

void SerialGPS::RcvIsrDisable(void)
{
#ifndef USE_ARDUINO
    ser.attach(NULL);
    Init();
#else  // USE_ARDUINO
#endif // USE_ARDUINO
}

///
/// private
///
#define NMEA_MAX_UNIT_SIZE (20)
#define NMEA_GGA_UNIT_SIZE (14)
#define NMEA_ZDA_UNIT_SIZE (6)

int32_t SerialGPS::ParseUnit(uint8_t *msg, int32_t len, uint8_t *unit[], int32_t max_unit)
{
    int32_t msg_cnt = 0;
    int32_t unit_cnt = 0;
    uint8_t *unit_ptr = &msg[0];

    for(msg_cnt = 0; msg_cnt <= len; msg_cnt++){
        if(msg[msg_cnt] == '\0'){
            // End of Message
            break;
        }
        if((msg[msg_cnt] == ',') || (msg[msg_cnt] == '*')){
            // Store the NULL character
            msg[msg_cnt] = '\0';
            // Store the start of unit pointer
            unit[unit_cnt] = unit_ptr;
            unit_cnt++;
            // Store the start of next unit pointer
            unit_ptr = &msg[msg_cnt + 1];
            // Check unit counter
            if(unit_cnt >= max_unit){
                break;
            }
        }
    }
    
    return unit_cnt;
}

int32_t SerialGPS::ParseGGA(uint8_t *msg, int32_t len, NMEA_GGA *nmea_gga)
{
    uint8_t *unit[NMEA_MAX_UNIT_SIZE];
    uint8_t unit_cnt;

    nmea_gga->error = 0x00000000;
    
    unit_cnt = ParseUnit(msg, len, unit, NMEA_MAX_UNIT_SIZE);
    if(unit_cnt != NMEA_GGA_UNIT_SIZE){
        // Unit size error
        nmea_gga->error |= 0x00000001;
        return -1;
    }
    
    // Time
    nmea_gga->hour = (ascii2dec(unit[0][0]) *  10) + (ascii2dec(unit[0][1]) *  1);
    nmea_gga->min  = (ascii2dec(unit[0][2]) *  10) + (ascii2dec(unit[0][3]) *  1);
    nmea_gga->sec  = (ascii2dec(unit[0][4]) *  10) + (ascii2dec(unit[0][5]) *  1);
    nmea_gga->msec = (ascii2dec(unit[0][7]) * 100) + (ascii2dec(unit[0][8]) * 10) + (ascii2dec(unit[0][9]) * 1);
    
    // Latitude
    if(sscanf((const char *)&unit[1][0], "%lf", &nmea_gga->latitude) != 1){
        // invalid latitude
        nmea_gga->latitude = 0.0;
        nmea_gga->error |= 0x00000002;
    }
    
    // Latitude unit
    nmea_gga->latitude_ns  = unit[2][0];
    if((nmea_gga->latitude_ns != 'N') && (nmea_gga->latitude_ns != 'S')){
        // invalid latitude unit
        nmea_gga->latitude_ns = '\0';
        nmea_gga->error |= 0x00000004;
    }
    
    // Longitude
    if(sscanf((const char *)&unit[3][0], "%lf", &nmea_gga->longitude) != 1){
        // invalid longitude
        nmea_gga->longitude = 0.0;
        nmea_gga->error |= 0x00000008;
    }
    
    // Longitude unit
    nmea_gga->longitude_ew  = unit[4][0];
    if((nmea_gga->longitude_ew != 'E') && (nmea_gga->longitude_ew != 'W')){
        // invalid longitude unit
        nmea_gga->longitude_ew = '\0';
        nmea_gga->error |= 0x00000010;
    }
    
    // Fix quality
    nmea_gga->fix_quality = atoi((const char *)&unit[5][0]);
    
    // Number of satellites being tracked
    if(strlen((const char *)&unit[6][0]) >= 1){
        nmea_gga->num_of_satellites = atoi((const char *)&unit[6][0]);
    } else {
        // invalid number of satellites being tracked
        nmea_gga->num_of_satellites = 0;
        nmea_gga->error |= 0x00000020;
    }
    
    // Horizontal dilution of position
    if(sscanf((const char *)&unit[7][0], "%lf", &nmea_gga->hdop) != 1){
        // invalid hdop
        nmea_gga->hdop = 0.0;
        nmea_gga->error |= 0x00000040;
    }

    // Altitude, Meters, above mean sea level
    if(sscanf((const char *)&unit[8][0], "%lf", &nmea_gga->altitude) != 1){
        // invalid altitude
        nmea_gga->altitude = 0.0;
        nmea_gga->error |= 0x00000080;
    }

    // Altitude unit
    nmea_gga->altitude_unit  = unit[9][0];
    if(nmea_gga->altitude_unit != 'M'){
        // invalid altitude unit
        nmea_gga->error |= 0x00000100;
    }
    
    // Height of geoid (mean sea level) above WGS84
    if(sscanf((const char *)&unit[10][0], "%lf", &nmea_gga->geoid_height) != 1){
        // invalid geoid height
        nmea_gga->geoid_height = 0.0;
        nmea_gga->error |= 0x00000200;
    }

    // Altitude unit
    nmea_gga->geoid_unit  = unit[11][0];
    if(nmea_gga->geoid_unit != 'M'){
        // invalid geoid height unit
        nmea_gga->error |= 0x00000400;
    }
    return 0;
}

int32_t SerialGPS::ParseZDA(uint8_t *msg, int32_t len, NMEA_ZDA *nmea_zda)
{
    uint8_t *unit[NMEA_MAX_UNIT_SIZE];
    uint8_t unit_cnt;
    
    nmea_zda->error = 0x00000000;
    
    unit_cnt = ParseUnit(msg, len, unit, NMEA_MAX_UNIT_SIZE);
    if(unit_cnt != NMEA_ZDA_UNIT_SIZE){
        // Unit size error
        nmea_zda->error |= 0x00000001;
        return -1;
    }
    
    // Time
    nmea_zda->hour = (ascii2dec(unit[0][0]) *   10) + (ascii2dec(unit[0][1]) *   1);
    nmea_zda->min  = (ascii2dec(unit[0][2]) *   10) + (ascii2dec(unit[0][3]) *   1);
    nmea_zda->sec  = (ascii2dec(unit[0][4]) *   10) + (ascii2dec(unit[0][5]) *   1);
    nmea_zda->msec = (ascii2dec(unit[0][7]) *  100) + (ascii2dec(unit[0][8]) *  10) + (ascii2dec(unit[0][9]) *  1);
    
    // Date
    nmea_zda->day  = (ascii2dec(unit[1][0]) *   10) + (ascii2dec(unit[1][1]) *   1);
    nmea_zda->month= (ascii2dec(unit[2][0]) *   10) + (ascii2dec(unit[2][1]) *   1);
    nmea_zda->year = (ascii2dec(unit[3][0]) * 1000) + (ascii2dec(unit[3][1]) * 100) + (ascii2dec(unit[3][2]) * 10) + (ascii2dec(unit[3][3]) * 1);
    
    // Time-zone difference
    if(strlen((const char *)&unit[4][0]) == 2){
        nmea_zda->hour_diff = (ascii2dec(unit[4][0]) * 10) + (ascii2dec(unit[4][1]) * 1);
    } else {
        nmea_zda->hour_diff = 0;
#ifdef GPS_CHECK_ERR        
        nmea_zda->error |= 0x00000002;
#endif
    }

    if(strlen((const char *)&unit[5][0]) == 2){
        nmea_zda->min_diff = (ascii2dec(unit[5][0]) * 10) + (ascii2dec(unit[5][1]) * 1);
    } else {
        nmea_zda->min_diff = 0;
#ifdef GPS_CHECK_ERR        
        nmea_zda->error |= 0x00000004;
#endif
    }

    return 0;
}

SerialGPS::RcvIsrState SerialGPS::GetRcvIsrState(void)
{
    return rcv_isr_state;
}

void SerialGPS::SetRcvIsrState(SerialGPS::RcvIsrState state)
{
    rcv_isr_state = state;
    
    if((rcv_isr_state == RCV_ISR_IDLE) ||
       (rcv_isr_state == RCV_ISR_READY)) {
        // Initialize NMEA message
        nmea_msg.Init();
        nmea_msg_sum = 0;
        nmea_msg_sum_fix = false;
    }
}

void SerialGPS::RcvIsrStateAidle(void)
{
#ifndef USE_ARDUINO
    uint8_t nmea_char = ser.getc();
#else  // USE_ARDUINO
    int nmea_char = Serial2.read();
    if(nmea_char == -1){
        return;
    }
#endif // USE_ARDUINO
    
    // Start of NMEA message?
    if(nmea_char == '$'){
        SetRcvIsrState(RCV_ISR_READY);
    }
}

void SerialGPS::RcvIsrStateReady(void)
{
#ifndef USE_ARDUINO
    uint8_t nmea_char = ser.getc();
#else  // USE_ARDUINO
    int nmea_char = Serial2.read();
    if(nmea_char == -1){
        return;
    }
#endif // USE_ARDUINO
    uint8_t sum = 0;

    // Store the character
    nmea_msg.msg[nmea_msg.len] = nmea_char;
    
    // End of checksum calculation?
    //  (Target is a character between the '*' from '$')
    if(nmea_char == '*'){
        nmea_msg_sum_fix = true;
    }

    // Calculate checksum
    if(nmea_msg_sum_fix == false){
        nmea_msg_sum ^= nmea_char;
    }

    // End of NMEA message?
    if( (nmea_char == '\r') || (nmea_char == '\n') ){
        // Store the NULL character
        nmea_msg.msg[nmea_msg.len] = '\0';
        nmea_msg.len++; // increment length

        // Check checksum
        sum  = ascii2hex(nmea_msg.msg[nmea_msg.len-3]) << 4;
        sum |= ascii2hex(nmea_msg.msg[nmea_msg.len-2]);
        if(sum == nmea_msg_sum){
            // Receiving completion.
            SetRcvIsrState(RCV_ISR_COMP);
        } else {
            // Checksum error. transition to idle.
            SetRcvIsrState(RCV_ISR_IDLE);
        }
        return;
    }
    
    nmea_msg.len++;
    if(nmea_msg.len >= QueueData::QUEUE_BUFFER_SIZE){
        // Buffer over run error. transition to idle.
        SetRcvIsrState(RCV_ISR_IDLE);
    }
}

void SerialGPS::RcvIsrHandler(void)
{
    switch(GetRcvIsrState()){
    case RCV_ISR_IDLE:
        RcvIsrStateAidle();
        break;
    case RCV_ISR_READY:
        RcvIsrStateReady();
        break;
    case RCV_ISR_COMP:
        // RcvIsrStateReady()で送信完了するのでここには来ない
    default:
        // Status error. transition to idle.
        SetRcvIsrState(RCV_ISR_IDLE);
        break;
    }
    
    // Processing of receive completion
    if(GetRcvIsrState() == RCV_ISR_COMP){
        // Enqueue
        queue.Enqueue(&nmea_msg);
        SetRcvIsrState(RCV_ISR_IDLE);
        return;
    }
}

#ifdef USE_ARDUINO
#endif

uint8_t SerialGPS::ascii2hex(uint8_t ascii)
{
    if((ascii >= '0') && (ascii <= '9')){
        return ascii - '0';
    } else if ((ascii >= 'A') && (ascii <= 'F')){
        return (ascii - 'A') + 0x0A;
    } else {
        return 0; // error
    }
}

uint8_t SerialGPS::ascii2dec(uint8_t ascii)
{
    if((ascii >= '0') && (ascii <= '9')){
        return ascii - '0';
    } else {
        return 0; // error
    }
}
