#ifndef _SERIAL_GPS_MODULE_
#define _SERIAL_GPS_MODULE_

#include "peer.h"
#ifdef USE_MBED
#include "mbed.h"
#else  // USE_MBED
#include <Arduino.h>
typedef int32_t PinName;
#endif // USE_MBED
#include "queue.h"

/**
 * Serial GPS(Global Positioning System) class.
 */
class SerialGPS {
public:
    /**
     * NMEA Sentence of GGA (Fix Information)
     */
    struct NMEA_GGA {
        // UTC Time (decimal)
        uint8_t hour;           // hour
        uint8_t min;            // minutes
        uint8_t sec;            // second
        uint16_t msec;          // milli second
        // Latitude
        double  latitude;       // latitude
        uint8_t latitude_ns;    // 'N'orth/'S'outh Latitude
        // Longitude
        double longitude;       // longitude
        uint8_t longitude_ew;   // 'E'ast/'W'est Longitude
        // Fix quality
        uint8_t fix_quality;
        // Number of satellites being tracked
        uint8_t num_of_satellites;
        // Horizontal dilution of position
        double hdop;
        // Altitude, Meters, above mean sea level
        double altitude;
        uint8_t altitude_unit;  // 'M'eter
        // Height of geoid (mean sea level) above WGS84
        double geoid_height;
        uint8_t geoid_unit;     // 'M'eter
        // Sentence error information (library original)
        uint32_t error;
    }nmea_gga;
    
    struct NMEA_GLL {}nmea_gll;
    struct NMEA_GSA {}nmea_gsa;
    struct NMEA_GSV {}nmea_gsv;
    struct NMEA_RMC {}nmea_rmc;
    struct NMEA_VTG {}nmea_vtg;

    /**
     * NMEA Sentence of ZDA (Date and Time)
     */
    struct NMEA_ZDA {
        // UTC Time (decimal)
        uint8_t hour;           // hour
        uint8_t min;            // minutes
        uint8_t sec;            // second
        uint16_t msec;          // milli second
        // Date
        uint8_t day;            // Day
        uint8_t month;          // Month
        uint16_t year;          // Year
        // Time-zone difference
        int8_t hour_diff;       // -13..13
        int8_t min_diff;        //   0..59
        // Sentence error information (library original)
        uint32_t error;
    }nmea_zda;
    
    /**
     * Callbak routine typedef
     */
    typedef void (*NMEA_LOG_CBFUNC)(uint8_t *);
    typedef void (*NMEA_GGA_CBFUNC)(NMEA_GGA *);
    typedef void (*NMEA_GLL_CBFUNC)(NMEA_GLL *);
    typedef void (*NMEA_GSA_CBFUNC)(NMEA_GSA *);
    typedef void (*NMEA_GSV_CBFUNC)(NMEA_GSV *);
    typedef void (*NMEA_RMC_CBFUNC)(NMEA_RMC *);
    typedef void (*NMEA_VTG_CBFUNC)(NMEA_VTG *);
    typedef void (*NMEA_ZDA_CBFUNC)(NMEA_ZDA *);

    struct GpsCallbackTable {
        NMEA_LOG_CBFUNC log_cbfunc;
        NMEA_GGA_CBFUNC gga_cbfunc;
        NMEA_GLL_CBFUNC gll_cbfunc;
        NMEA_GSA_CBFUNC gsa_cbfunc;
        NMEA_GSV_CBFUNC gsv_cbfunc;
        NMEA_RMC_CBFUNC rmc_cbfunc;
        NMEA_VTG_CBFUNC vtg_cbfunc;
        NMEA_ZDA_CBFUNC zda_cbfunc;
    };
    
    /**
     * Constructor
     *
     * @param tx A pin of transmit.
     * @param rx A pin of recieve.
     * @param baud Baud rate. (Default = 9600)
     */
    SerialGPS(PinName tx, PinName rx, int baud = 9600);
    
    /**
     * Destructor
     */
    ~SerialGPS();
    
    /**
     * Initialize
     */
    void Init(void);
    
    /**
     * GPS Message Processing (Forever mode)
     */
    void Processing(void);
    
    /**
     * GPS Message Processing (One-shot mode)
     *
     * @return Processing continues?
     */
    bool ProcessingOneShot(void);

    /**
     * Set a callback table
     *
     * @param table A table of callback function
     */
    void SetGpsCallback(GpsCallbackTable *table);

    /**
     * GPS Message receive enable
     */
    void RcvIsrEnable(void);
    
    /**
     * GPS Message receive Disable
     */
    void RcvIsrDisable(void);

    /**
     * Receiving interrupt handler
     */
    void RcvIsrHandler(void);
private:
    // Receiving interrupt status
    enum RcvIsrState {
        RCV_ISR_IDLE  = 0, // Idle 
        RCV_ISR_READY = 1, // Receiving
        RCV_ISR_COMP  = 2  // Receiving completion
    };
    RcvIsrState rcv_isr_state;

    // NMEA sentences
    #define GPS_NMEA_LEN (5)
    #define GPS_NMEA_GPGGA ("GPGGA") // GGA: Fix information
    #define GPS_NMEA_GPGLL ("GPGLL") // GLL: Lat/Lon data
    #define GPS_NMEA_GPGSA ("GPGSA") // GSA: Overall Satellite data
    #define GPS_NMEA_GPGSV ("GPGSV") // GSV: Detailed Satellite data
    #define GPS_NMEA_GPRMC ("GPRMC") // RMC: recommended minimum data for gps
    #define GPS_NMEA_GPVTG ("GPVTG") // VIG: Vector track an Speed over the Ground
    #define GPS_NMEA_GPZDA ("GPZDA") // ZDA: Date and Time
    
    // GPS callback function
    GpsCallbackTable *gps_callback_table;
    
    // GPS message queue
    Queue queue;
    // GPS message buffer
    QueueData nmea_msg;      // Message
    uint8_t nmea_msg_sum;    // Checksum
    bool nmea_msg_sum_fix;   // Fix Checksum flag

    // Serial communication object
#ifdef USE_MBED
    Serial ser;
#endif
    
    /**
     * Parsing NMEA sentence unit
     */
    int32_t ParseUnit(uint8_t *msg, int32_t len, uint8_t *unit[], int32_t max_unit);

    /**
     * Parsing NMEA GGA sentence
     *
     * @param msg A NMEA message pointer
     * @param len A length of NEMA messege pointer
     * @param nmea_gga A parse result
     * @return error (0: OK, -1:NG)
     */
    int32_t ParseGGA(uint8_t *msg, int32_t len, NMEA_GGA *nmea_gga);
    
    /**
     * Parsing NMEA ZDA sentence
     *
     * @param msg A NMEA message pointer
     * @param len A length of NEMA messege pointer
     * @param nmea_zda A parse result
     * @return error (0: OK, -1:NG)
     */
    int32_t ParseZDA(uint8_t *msg, int32_t len, NMEA_ZDA *nmea_zda);

    /**
     * Get a receiving interrupt status
     *
     * @return Receiving interrupt status
     */
    RcvIsrState GetRcvIsrState(void);

    /**
     * Set a receiving interrupt status
     *
     * @param state A receiving interrupt status
     *
     * When 'IDLE' or 'READY' is set as 'state', receiving condition is initialized.
     */
    void SetRcvIsrState(RcvIsrState state);

    /**
     * Receiving interrupt handler for IDLE status
     */
    void RcvIsrStateAidle(void);

    /**
     * Receiving interrupt handler for READY status
     */
    void RcvIsrStateReady(void);

    /**
     * Convert from ASCII to HEX
     * 
     * @param ascii A ASCII character ('0' to '9' and 'A' to 'F')
     * @return Convert result (HEX)
     */
    uint8_t ascii2hex(uint8_t ascii);

    /**
     * Convert from ASCII to DEC
     * 
     * @param ascii A ASCII character ('0' to '9')
     * @return Convert result (DEC)
     */
    uint8_t ascii2dec(uint8_t ascii);
};

#endif 